// fdm86 - cochlear model solutions described in JASA 79, 1472-1480

#include <stdio.h>
#include <math.h>

#define Nmax 256 

typedef struct {double x, y;} complex;

complex a[Nmax], p[Nmax], y[Nmax], ame, pme, s, srd;
complex dbm[Nmax], dhc[Nmax];
double dx, dy, ped, dst;

int m = 1;
int n = 251;
double xl = 2.5;
double yw = 0.1;
double zh = 0.1;
double rho = 1;
double rhe = 0;

/* set of complex arithmetic functions */

complex
c_add(complex z1, complex z2)		/* complex add */
{
    complex z3;

    z3.x = z1.x + z2.x;
    z3.y = z1.y + z2.y;

    return (z3);
}

complex
c_sub(complex z1, complex z2)		/* complex subtract */
{
    complex z3;

    z3.x = z1.x - z2.x;
    z3.y = z1.y - z2.y;

    return (z3);
}

complex
c_mul(complex z1, complex z2)		/* complex multiply */
{
    complex z3;

    z3.x = z1.x * z2.x - z1.y * z2.y;
    z3.y = z1.y * z2.x + z1.x * z2.y;

    return (z3);
}

complex
c_div(complex z1, complex z2)		/* complex divide */
{
    complex z3;
    double ms;

    ms = z2.x * z2.x + z2.y * z2.y;
    z3.x = (z1.x * z2.x + z1.y * z2.y) / ms;
    z3.y = (z1.y * z2.x - z1.x * z2.y) / ms;

    return (z3);
}

complex
c_rcp(complex z)			/* complex reciprocal */
{
    double ms;

    ms = z.x * z.x + z.y * z.y;
    z.x = z.x / ms;
    z.y = -z.y / ms;

    return (z);
}

complex
c_ri(double zr, double zi)		/* complex (real,imaginary) */
{
    complex z;

    z.x = zr;
    z.y = zi;

    return (z);
}

/* c_krm - compute impedance z(f) = k/(iw) + r + m*(iw)
 *          where w=2*pi*f
 */
complex
c_krm(double k, double r, double m, double f)
{
    complex z;
    double w;
    static double tpi = 2 * 3.1415927;

    w = tpi * f;
    z.x = r;
    z.y = m * w - k / w;

    return (z);
}

/* c_krmx - compute impedance z(f,x) = k(x)/(iw) + r(x) + m(x)*(iw) 
 *          where w=2*pi*f
 *                k(x)=k0*exp(ke*x)
 *                r(x)=r0*exp(re*x)
 *                m(x)=m0*exp(me*x)
 */
complex
c_krmx(double ko, double ke, double ro, double re, double rc,
    double mo, double me, double f, double x)
{
    complex z;
    double w;
    static double tpi = 2 * 3.1415927;

    w = tpi * f;
    z.x = ro * exp (re * x) + rc;
    z.y = mo * exp (me * x) * w - ko * exp (ke * x) / w;

    return (z);
}

/* admit - compute admittance
 *         mode=0  =>  partition driving-point admittance
 *         mode=1  =>  BM admittance
 *         mode=2  =>  HC admittance
 */
void
admit(double f, int mode)
{
    complex z1, z2, z3, z4, detz, zme, t1, t2, t3, zb, hc, za;
    complex srd, srdbp, bgsrdbp;
    double td, x, bw, gp, gg, aa;
    int i;
    static double tpi = 2 * 3.1415927;
    static double kst = 5.e5;               /* stiffness of stapes attach. */
    static double rst = 80;                 /* damping of stapes */
    static double mst = 0.005;              /* eff. mass of stapes */
    static double ast = 0.01;               /* eff. area stapes footplate */
    static double kma = 4.e5;               /* stiffness of malleus attach. */
    static double rma = 80;                 /* damping of malleus */
    static double mma = 0.01;               /* eff. mass of malleus */
    static double ama = 0.35;               /* eff. area  eardrum */
    static double kim = 8.e7;               /* incudo-malleolar stiffness */
    static double rim = 800;                /* incudo-malleolar damping */
    static double gmel = 0.5;               /* middle ear lever gain */
    static double bwo = 0.04, bwe = 0;	    /* BM width */
    static double gpo = 1, gpe = 0;	    /* BM to IHC lever gain */
    static double gam = 1;		    /* normal OHC transduction gain */
    static double k1o = 1.1e9, k1e = -4;    /* stiffness component of Z1 */
    static double r1o = 1500, r1e = -2;     /* damping component of Z1 */
    static double r1c = 20;		    /* r1(x) = r1o*exp(r1e*x) + r1c */
    static double m1o = 0.003, m1e = 0;     /* mass component of Z1 */
    static double k2o = 7e6, k2e = -4.4;    /* stiffness component of Z2 */
    static double r2o = 10, r2e = -2.2;     /* damping component of Z2 */
    static double r2c = 0;                  /* r2(x) = r2o*exp(r2e*x) + r2c */
    static double m2o = 0.0005, m2e = 0;    /* mass component of Z2 */
    static double k3o = 1e7, k3e = -4;      /* stiffness component of Z3 */
    static double r3o = 2, r3e = -0.8;      /* damping component of Z3 */
    static double k4o = 6.15e8, k4e = -4;   /* stiffness component of Z4 */
    static double r4o = 1040, r4e = -2;     /* damping component of Z4 */

    dx = xl / (n - 1);
    s.x = 0;
    s.y = tpi * f;
    srd.x = 0;
    srd.y = 2*s.y*rho*dx;
    srdbp.x = 0;
    srdbp.y = (0.5 / yw) * (srd.y * dx / zh);

    /* Cochlear partition */
    for(i = 0; i < n; i++) {
	x = i * dx;
	z1 = c_krmx(k1o, k1e, r1o, r1e, r1c, m1o, m1e, f, x);
	z2 = c_krmx(k2o, k2e, r2o, r2e, r2c, m2o, m2e, f, x);
	z3 = c_krmx(k3o, k3e, r3o, r3e, 0, 0, 0, f, x);
	z4 = c_krmx(k4o, k4e, r4o, r4e, 0, 0, 0, f, x);
	bw = bwo*exp(bwe*x);
	gp = gpo*exp(gpe*x);
	/* bgsrdbp = (bw / gp) * srdbp */
	bgsrdbp = c_ri(0, (bw / gp) * srdbp.y);
	/* hc = z2 / (z2 + z3) */
	hc = c_div(z2, c_add(z2, z3));
	/* za = hc * (z3 - gam * z4) */
	za = c_mul(hc, c_sub(z3, c_mul(c_ri(gam, 0), z4)));
	/* zb = z1 + za */
	zb = c_add(z1, za);
	if (mode == 0) {	    /* partition driving-point admittance */
	    /* y(i) = bgsrdbp / zb */
	    y[i] = c_div(bgsrdbp, zb);
	} else if (mode == 1) {	    /* bm admittance */
	    /* y(i) = 1 / zb */
	    y[i] = c_rcp(zb);
	} else if (mode == 2) {	    /* hc admittance */
	    /* y(i) = hc / zb */
	    y[i] = c_div(hc, zb);
	}
    }

    /* Middle-ear */

    ped = 2.828e-4;     /* eardrum pressure (dyn/cm**2 peak) 
		           = 20 micropascal rms = 0 dB SPL. */
    gg = gmel*gmel;
    z1 = c_krm(kma, rma, mma, f);
    z2 = c_krm(kst, rst, mst, f);
    z3 = c_krm(kim, rim, 0, f);
    /* detz = z1 * z2 + z2 * z3 + gg * z3 * z1 */
    t1 = c_mul(z1, z2);
    t2 = c_mul(z2, z3);
    t3 = c_mul(z3, z1);
    detz.x = t1.x + t2.x + gg * t3.x;
    detz.y = t1.y + t2.y + gg * t3.y;
    /* zme  =  (z1 / gg + z2) / ast */
    zme.x  =  (z1.x / gg + z2.x) / ast;
    zme.y  =  (z1.y / gg + z2.y) / ast;
    /* ame  =  -(srd * ast) / (yw * zh * zme) */
    aa  = -ast / (yw * zh);
    td = zme.x * zme.x + zme.y * zme.y;
    ame.x  =  aa * (srd.y * zme.y) / td;
    ame.y  =  aa * (srd.y * zme.x) / td;
    /* pme  =  ame * ped * ama / (ast * gmel) */
    pme.x  =  ame.x * ped * (ama / (ast * gmel));
    pme.y  =  ame.y * ped * (ama / (ast * gmel));
}

/* dbph - compute magnitude (in dB) and phase (unwrapped) */

void
dbph(complex *p, int n)
{
    int i;
    double mg, db, ph, lp = -0.5;
    static double tpi = 2 * 3.1415927;
    
    for (i = 0; i < n; i++) {
        mg = hypot(p[i].x, p[i].y);
        if (mg > 1e-30) {
            db = 20 * log10(mg);
            ph = atan2(p[i].y, p[i].x) / tpi;
        } else {
            db = -600;
            ph = 0;
        }
        if (ph > lp) {
            ph -= floor(ph - lp + 0.5);
        } else {
            ph += floor(lp - ph + 0.5);
        }
        lp = ph;
        p[i].x = db;
        p[i].y = ph;
    }
}

/* wrsoln - write pressure, BM disp, and HC disp to file */

void
wrsoln(int ii, double f)
{
    char fn[80];
    complex tc;
    double pr, dr, dl, ps;
    int i;
    FILE *fp;

    sprintf(fn, "fdm86_%d.dat", ii);
    fp = fopen(fn, "w");
    fprintf(fp, "; %s - f=%.0f\n", fn, f);
    fprintf(fp, "; m=%d n=%d xl=%.2f yw=%.2f zh=%.2f\n", m, n, xl, yw, zh);

    /* BM displacement */
    admit(f, 1);
    for (i = 0; i < n; i++) {
    	tc.x = y[i].x * p[i].x - y[i].y * p[i].y;
    	tc.y = y[i].x * p[i].y + y[i].y * p[i].x;
    	dbm[i].x = tc.y / s.y;
    	dbm[i].y = -tc.x / s.y;
    }

    /* HC displacement */
    admit(f, 2);
    for (i = 0; i < n; i++) {
    	tc.x = y[i].x * p[i].x - y[i].y * p[i].y;
    	tc.y = y[i].x * p[i].y + y[i].y * p[i].x;
    	dhc[i].x = tc.y / s.y;
    	dhc[i].y = -tc.x / s.y;
    }

    /* convert to magnitude and phase */
    dbph(p, n);
    dbph(dbm, n);
    dbph(dhc, n);

    /* write data */
    fprintf(fp, "; x/L          P                 Dbm               Dhc     \n");
    fprintf(fp, ";         [dB]     [ph]     [dB]     [ph]     [dB]     [ph] \n");
    pr = 20 * log10(ped) + 3; /* ??? */
    dr = 20 * log10(1e-7);
    dl = 1.0 / (n - 1);
    ps = pr - 300;
    for (i = 0; i < n && p[i].x > ps; i++) {
        fprintf(fp, "%6.3f %8.1f %8.2f %8.1f %8.2f %8.1f %8.2f\n", i * dl,
        p[i].x - pr, p[i].y, dbm[i].x - dr, dbm[i].y, dhc[i].x - dr, dhc[i].y);
    }
    fclose(fp);
}

/* fdmodl - compute 1-D fluid pressure */

void
fdmodl(double f)
{
    complex tc;
    double td;
    int k;

    /* rigid basal boundary */
    /* a(1) = 1 / (1 - ame) */
    tc.x = 1 - ame.x;
    tc.y = -ame.y;
    td = tc.x * tc.x + tc.y * tc.y;
    a[0].x = tc.x / td;
    a[0].y = -tc.y / td;
    /* p(1) = pme * a(1) */
    p[0].x = pme.x * a[0].x - pme.y * a[0].y;
    p[0].y = pme.x * a[0].y + pme.y * a[0].x;

    /* from base to apex */
    for(k = 1; k < n - 1; k++) {
	/* a(k) = 1 / (2 + y(k) - a(k-1)) */
	tc.x = 2 + y[k].x - a[k - 1].x;
	tc.y = y[k].y - a[k-1].y;
	td = tc.x * tc.x + tc.y * tc.y;
	a[k].x = tc.x / td;
	a[k].y = -tc.y / td;
	/* p(k) = p(k-1) * a(k) */
	p[k].x = p[k-1].x * a[k].x - p[k-1].y * a[k].y;
	p[k].y = p[k-1].x * a[k].y + p[k-1].y * a[k].x;
    }

    /* apex */
    /* a(n) = rhe / (rhe*(1 - a(n-1)) + 2*s*rho*dx) */
    tc.x = rhe * (1 - a[n - 2].x);
    tc.y = 2 * s.y * rho * dx - rhe * a[n-2].y;
    td = tc.x * tc.x + tc.y * tc.y;
    a[n-1].x = rhe * tc.x / td;
    a[n-1].y = -rhe * tc.y / td;
    /* p(n) = p(n-1)*a(n) */
    p[n-1].x = p[n-2].x * a[n-1].x - p[n-2].y * a[n-1].y;
    p[n-1].y = p[n-2].x * a[n-1].y + p[n-2].y * a[n-1].x;

    /* from apex to base */
    for(k = n-2; k >= 0; k--) {
	/* p(k) = p(k) + a(k)*p(k+1) */
	p[k].x += a[k].x * p[k+1].x - a[k].y * p[k+1].y;
	p[k].y += a[k].x * p[k+1].y + a[k].y * p[k+1].x;
    }

}


int
main()
{
    double f;
    int i;

     for (i = 1, f = 400; f <= 25600; i++, f *= 4) {
        admit(f, 0);	/* define driving-point impedance */
	fdmodl(f);	/* solve 1-D fluid pressure */
	wrsoln(i, f);	/* write pressure, BM disp, and HC disp */
    }
    
     return (0);
}
